#include "tokens.h"

#include <assert.h>
#include <stdbool.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>

const char *TokenType_to_str(const enum TokenType token_type)
{
    switch (token_type) {
    case TOKEN_ROUND_OPEN:    return "TOKEN_ROUND_OPEN";
    case TOKEN_ROUND_CLOSE:   return "TOKEN_ROUND_CLOSE";
    case TOKEN_SQUARE_OPEN:   return "TOKEN_SQUARE_OPEN";
    case TOKEN_SQUARE_CLOSE:  return "TOKEN_SQUARE_CLOSE";
    case TOKEN_CURLY_OPEN:    return "TOKEN_CURLY_OPEN";
    case TOKEN_CURLY_CLOSE:   return "TOKEN_CURLY_CLOSE";
    case TOKEN_QUOTE:         return "TOKEN_QUOTE";
    case TOKEN_QUASI_QUOTE:   return "TOKEN_QUASI_QUOTE";
    case TOKEN_QUASI_UNQUOTE: return "TOKEN_QUASI_UNQUOTE";
    case TOKEN_TAG:           return "TOKEN_TAG";
    case TOKEN_IDENT:         return "TOKEN_IDENT";
    case TOKEN_NUM:           return "TOKEN_NUM";
    case TOKEN_STRING:        return "TOKEN_STRING";
    }
    return NULL;
}

Tokens Tokens_new()
{
    Tokens tokens = malloc(sizeof(struct Tokens));
    assert(tokens);
    memset(tokens, 0, sizeof(struct Tokens));
    tokens->top = NULL;
    return tokens;
}

void Tokens_delete(const Tokens tokens)
{
    assert(tokens);
    while (tokens->top) Tokens_pop(tokens);
    free(tokens);
}

Token Tokens_top(const Tokens self)
{
    return self->top;
}

bool Tokens_expect(const Tokens self, const enum TokenType token_type)
{
    if (!self->top || self->top->type != token_type) return false;
    Tokens_pop(self);
    return true;
}

void Tokens_pop(const Tokens self)
{
    if (self->top) {
        const Token token = self->top;
        self->top = token->next;
        if (token->val) free(token->val);
        free(token);
    }
}

void Tokens_append(const Tokens self,
                   const enum TokenType token_type, const char *const val)
{
    const Token token = malloc(sizeof(struct Token));
    assert(token);
    token->next = NULL;
    token->type = token_type;
    token->val = malloc(strlen(val) + 1);
    assert(token->val);
    strcpy(token->val, val);

    if (!self->top) {
        self->top = token;
    } else for (Token curr = self->top; curr; curr = curr->next) {
        if (!curr->next) {
            curr->next = token;
            break;
        }
    }
}
