#ifndef __ARCANA_LISP_TOKENS_H__
#define __ARCANA_LISP_TOKENS_H__

#include <stdbool.h>
#include <stddef.h>

#define TOKENS_DELETE(tokens) do { \
    Tokens_delete(tokens);         \
    tokens = NULL;                 \
} while (0)

enum TokenType {
    TOKEN_ROUND_OPEN,
    TOKEN_ROUND_CLOSE,
    TOKEN_SQUARE_OPEN,
    TOKEN_SQUARE_CLOSE,
    TOKEN_CURLY_OPEN,
    TOKEN_CURLY_CLOSE,
    TOKEN_QUOTE,
    TOKEN_QUASI_QUOTE,
    TOKEN_QUASI_UNQUOTE,
    TOKEN_TAG,
    TOKEN_IDENT,
    TOKEN_NUM,
    TOKEN_STRING,
};

typedef struct Tokens {
    struct Token *top;
} *Tokens;

typedef struct Token {
    struct Token *next;
    enum TokenType type;
    char *val;
} *Token;

const char *TokenType_to_str(enum TokenType token_type);

Tokens Tokens_new();
void Tokens_delete(Tokens self);

void Tokens_append(Tokens self, enum TokenType token_type, const char *val);
void Tokens_pop(Tokens self);
bool Tokens_expect(Tokens self, enum TokenType token_type);
Token Tokens_top(Tokens self);

#endif
